package org.artifactory.ph;

import org.artifactory.api.repo.exception.BlackedOutException;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.repo.exception.SnapshotPolicyException;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.DownloadRequestContext;
import org.artifactory.resource.BlockedRepoResource;
import org.artifactory.resource.RemoteRepoResource;
import org.artifactory.resource.UnfoundRepoResource;
import org.artifactory.sapi.fs.VfsFile;

import javax.validation.constraints.NotNull;
import java.io.IOException;

import static org.apache.http.HttpStatus.SC_CONFLICT;
import static org.apache.http.HttpStatus.SC_FORBIDDEN;
import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.apache.http.HttpStatus.SC_NOT_FOUND;
import static org.apache.http.HttpStatus.SC_UNAUTHORIZED;

public class ArtifactExtractor {

    private static final String NOT_FOUND = "Not found";

    private ArtifactExtractor() {
    }

    @NotNull
    public static VfsFile<?> locateFile(DownloadRequestContext context,
            InternalRepositoryService repositoryService) throws PackageException {
        Repo<?> repo = repositoryService.repositoryByKey(context.getRepoKey());
        if (repo == null) {
            throw repoNotFoundException(context.getRepoKey(), context.getPath());
        }
        RepoResource info = repo.getInfo(context);
        return locateAndGetByInfo(info, context, repositoryService);
    }

    @NotNull
    private static VfsFile<?> locateAndGetByInfo(RepoResource resource, DownloadRequestContext context,
            InternalRepositoryService repositoryService) throws PackageException {
        if (resource.isFound()) {
            if (resource instanceof RemoteRepoResource) {
                Repo<?> repo = repositoryService.remoteRepositoryByKey(resource.getRepoKey());
                fillRemoteRepoCache(resource, context, repo, repositoryService);
            }
            VfsFile<?> file = locateFileInternal(resource, repositoryService);
            if (file == null) {
                file = locateFileInternal(resource, repositoryService);
                if (file == null) {
                    throw internalErrorArtifactExpected(resource.getRepoKey(), resource.getPath());
                }
            }
            return file;
        } else {
            throwException(resource);
        }
        throw artifactNotFoundException(resource.getRepoKey(), resource.getPath());
    }

    @NotNull
    private static void throwException(RepoResource info) throws PackageException {
        if (info instanceof UnfoundRepoResource) {
            UnfoundRepoResource res = (UnfoundRepoResource) info;
            switch (res.getStatusCode()) {
                case SC_CONFLICT:
                    throw conflictException(info.getRepoKey(), info.getPath(), res.getDetail());
                case SC_FORBIDDEN:
                    throw noAuthorization(info.getRepoKey(), info.getPath());
                default:
                    throw artifactNotFoundException(info.getRepoKey(), info.getPath());
            }
        }
        if (info instanceof BlockedRepoResource) {
            throw blockedException(info.getRepoKey(), info.getPath());
        }
    }

    private static void fillRemoteRepoCache(RepoResource info, DownloadRequestContext context, Repo<?> repo,
            InternalRepositoryService repositoryService)
            throws PackageException {
        try {
            repositoryService.getResourceStreamHandle(context, repo, info);
        } catch (SnapshotPolicyException e) {
            throw conflictException(info.getRepoKey(), info.getPath(), e.getMessage());
        } catch (BlackedOutException e) {
            throw blackedOuttFoundException(info.getRepoKey(), info.getPath());
        } catch (RepoRejectException e) {
            throw rejectException(info.getRepoKey(), info.getPath(), e.getMessage());
        } catch (IOException e) {
            throw internalError(info.getRepoKey(), info.getPath(), e);
        }
    }

    private static VfsFile<?> locateFileInternal(RepoResource info, InternalRepositoryService repositoryService) {
        LocalRepo<?> localRepo = repositoryService.localFederatedOrCachedRepositoryByKey(info.getRepoKey());
        if (localRepo != null) {
            return localRepo.getImmutableFile(RepoPathFactory.create(localRepo.getKey(), info.getPath()));
        }
        return null;
    }

    private static PackageException repoNotFoundException(String repoKey, String path) {
        String message = String.format("Failed to find repository in repo:%s, path:%s", repoKey, path);
        return new PackageException(SC_NOT_FOUND, NOT_FOUND, message);
    }

    private static PackageException blackedOuttFoundException(String repoKey, String path) {
        String message = String.format("Failed to find artifact in repo:%s, path:%s, cause: repo is blacked out",
                repoKey, path);
        return new PackageException(SC_NOT_FOUND, NOT_FOUND, message);
    }

    private static PackageException blockedException(String repoKey, String path) {
        String message = String.format("Failed to find repository in repo:%s, path:%s, cause: artifact blocked",
                repoKey, path);
        return new PackageException(SC_NOT_FOUND, NOT_FOUND, message);
    }

    private static PackageException artifactNotFoundException(String repoKey, String path) {
        String message = String.format("Failed to find artifact in repo:%s, path:%s, cause: not found", repoKey, path);
        return new PackageException(SC_NOT_FOUND, NOT_FOUND, message);
    }

    private static PackageException conflictException(String repoKey, String path, String cause) {
        String message = String.format("Failed to find artifact in repo:%s, path:%s, cause:%s", repoKey, path, cause);
        return new PackageException(SC_CONFLICT, "Conflict", message);
    }

    private static PackageException noAuthorization(String repoKey, String path) {
        String message = String.format("User doesn't have the required authorization to access artifact" +
                " in repo:%s, path:%s", repoKey, path);
        return new PackageException(SC_UNAUTHORIZED, "Unauthorized", message);
    }

    private static PackageException rejectException(String repoKey, String path, String cause) {
        String message = String.format("Failed to find repository in repo:%s, path:%s, cause:%s", repoKey, path, cause);
        return new PackageException(SC_NOT_FOUND, NOT_FOUND, message);
    }

    private static PackageException internalErrorArtifactExpected(String repoKey, String path) {
        String message = String.format("Failed to find artifact, repo:%s, path:%s, " +
                "cause: artifact should not be null (resource found)", repoKey, path);
        return new PackageException(SC_INTERNAL_SERVER_ERROR, "Internal server error", message);
    }

    private static PackageException internalError(String repoKey, String path, Exception e) {
        String message = String.format("Failed to find artifact, repo:%s, path:%s, cause:%s", repoKey, path, e);
        return new PackageException(SC_INTERNAL_SERVER_ERROR, "Internal server error", message, e);
    }
}
