package org.artifactory.projects.interceptor;

import org.artifactory.config.ConfigurationChangesInterceptor;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.exception.AccessClientStateException;
import org.artifactory.repo.RepoDetailsType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Map;

/**
 * Handles repository removal from Projects (managed in Access) for local, remote and virtual repositories
 *
 * @author andreik
 */
@Component
public class ProjectRepoConfigurationInterceptor implements ConfigurationChangesInterceptor {

    private static final Logger log = LoggerFactory.getLogger(ProjectRepoConfigurationInterceptor.class);

    private final InternalCentralConfigService internalCentralConfigService;
    private final ProjectsService projectsService;

    @Autowired
    public ProjectRepoConfigurationInterceptor(InternalCentralConfigService internalCentralConfigService,
            ProjectsService projectsService) {
        this.internalCentralConfigService = internalCentralConfigService;
        this.projectsService = projectsService;
    }

    @Override
    public void onBeforeSave(CentralConfigDescriptor newDescriptor) {
        CentralConfigDescriptor oldDescriptor = internalCentralConfigService.getDescriptor();
        if (oldDescriptor != null && newDescriptor != null) {
            Map<String, ? extends RepoDescriptor> newRepos = newDescriptor.getLocalRepositoriesMap();
            Map<String, ? extends RepoDescriptor> oldRepos = oldDescriptor.getLocalRepositoriesMap();
            removeDeletedRepositoriesFromProjects(newRepos, oldRepos, RepoDetailsType.LOCAL);

            newRepos = newDescriptor.getFederatedRepositoriesMap();
            oldRepos = oldDescriptor.getFederatedRepositoriesMap();
            removeDeletedRepositoriesFromProjects(newRepos, oldRepos, RepoDetailsType.FEDERATED);

            newRepos = newDescriptor.getRemoteRepositoriesMap();
            oldRepos = oldDescriptor.getRemoteRepositoriesMap();
            removeDeletedRepositoriesFromProjects(newRepos, oldRepos, RepoDetailsType.REMOTE);

            newRepos = newDescriptor.getVirtualRepositoriesMap();
            oldRepos = oldDescriptor.getVirtualRepositoriesMap();
            removeDeletedRepositoriesFromProjects(newRepos, oldRepos, RepoDetailsType.VIRTUAL);
        }
    }

    private void removeDeletedRepositoriesFromProjects(Map<String, ? extends RepoDescriptor> newRepos,
            Map<String, ? extends RepoDescriptor> oldRepos, RepoDetailsType repoType) {
        if (newRepos != null && oldRepos != null) {
            for (Map.Entry<String, ? extends RepoDescriptor> repoKeyEntry : oldRepos.entrySet()) {
                String repoKey = repoKeyEntry.getKey();
                if (!newRepos.containsKey(repoKey)) {
                    try {
                        // for every project remove the repo
                        log.debug("Repo '{}' from type '{}' was deleted, removing from Access", repoKey, repoType);
                        projectsService.removeRepositoryFromCurrentProject(repoKey);
                    } catch (AccessClientStateException e) {
                        log.warn("Failed to remove repo '{}' from a project association: {}. Will be handled by " +
                                "SyncResourcesWithAccessJob", repoKey, e.getMessage());
                    } catch (Exception e) {
                        log.error("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage());
                        log.debug("Failed to remove repo '{}' from a project association: {}", repoKey, e.getMessage(),
                                e);
                    }
                }
            }
        }
    }
}
