package org.artifactory.probes.service;

import com.jfrog.sysconf.SysConfig;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.probes.model.ReadinessStatus;
import org.artifactory.schedule.TaskService;
import org.artifactory.security.access.AccessService;
import org.artifactory.state.ArtifactoryServerState;
import org.artifactory.storage.StorageService;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.util.ReflectionUtils;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.ws.rs.ServiceUnavailableException;
import java.io.File;
import java.lang.reflect.Field;

import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertThrows;
import static org.testng.Assert.assertTrue;

/**
 * @author Andrey Tashlitsky
 */
public class ProbesServiceImplTest {

    @Mock
    private ArtifactoryServersCommonService artifactoryServers;
    @Mock
    ArtifactoryServer artifactoryServer;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private SysConfig sysConfig;
    @Mock
    private File dataDir;
    @Mock
    private File logDir;
    @Mock
    private StorageService storageService;
    @Mock
    private InternalDbService internalDbService;
    @Mock
    private AccessService accessService;
    @Mock
    private TaskService taskService;

    private AutoCloseable autoCloseable;
    private ProbesService probesService;

    @BeforeMethod
    public void beforeMethod() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ArtifactoryHome.bind(artifactoryHome);
    }

    @AfterMethod
    public void afterMethod() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
        autoCloseable.close();
    }

    @Test(dataProvider = "testTriggerReadinessChecksDataProvider")
    public void testTriggerReadinessChecks(ArtifactoryServerState serverState, boolean offline, boolean canWriteDataDir,
            boolean canWriteLogDir, boolean expectedIsReady) {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(serverState);
        when(artifactoryContext.isOffline()).thenReturn(offline);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(canWriteDataDir);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(canWriteLogDir);

        doNothing().when(storageService).ping(true);
        doNothing().when(internalDbService).health();
        doNothing().when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = (ReadinessStatus) ReflectionUtils.getField(f, probesService);
        assertNotNull(readinessStatus);
        assertEquals(readinessStatus.isReady(), expectedIsReady);
    }

    @Test
    public void testTriggerReadinessChecksLastSuccessfulReadinessNotZero() {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(ArtifactoryServerState.RUNNING);
        when(artifactoryContext.isOffline()).thenReturn(false);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(true);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(true);

        doNothing().when(storageService).ping(true);
        doNothing().when(internalDbService).health();
        doNothing().when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "lastSuccessfulReadiness");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        Long lastSuccessfulReadiness = (Long) ReflectionUtils.getField(f, probesService);
        assertNotNull(lastSuccessfulReadiness);
        assertNotEquals(lastSuccessfulReadiness, 0L);
    }

    @Test(dataProvider = "testTriggerReadinessChecksLastSuccessfulReadinessZeroDataProvider")
    public void testTriggerReadinessChecksLastSuccessfulReadinessZero(ArtifactoryServerState serverState,
            boolean offline, boolean canWriteDataDir, boolean canWriteLogDir) {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(serverState);
        when(artifactoryContext.isOffline()).thenReturn(offline);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(canWriteDataDir);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(canWriteLogDir);

        doNothing().when(storageService).ping(true);
        doNothing().when(internalDbService).health();
        doNothing().when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "lastSuccessfulReadiness");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        Long lastSuccessfulReadiness = (Long) ReflectionUtils.getField(f, probesService);
        assertNotNull(lastSuccessfulReadiness);
        assertEquals((long) lastSuccessfulReadiness, 0L);
    }

    @Test
    public void testTriggerReadinessChecksStorageServiceThrowsException() {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(ArtifactoryServerState.RUNNING);
        when(artifactoryContext.isOffline()).thenReturn(false);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(true);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(true);

        doThrow(new RuntimeException("What are you doing?")).when(storageService).ping(true);
        doNothing().when(internalDbService).health();
        doNothing().when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = (ReadinessStatus) ReflectionUtils.getField(f, probesService);
        assertNotNull(readinessStatus);
        assertFalse(readinessStatus.isReady());
    }

    @Test
    public void testTriggerReadinessChecksInternalDbServiceThrowsException() {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(ArtifactoryServerState.RUNNING);
        when(artifactoryContext.isOffline()).thenReturn(false);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(true);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(true);

        doNothing().when(storageService).ping(true);
        doThrow(new RuntimeException("What are you doing?")).when(internalDbService).health();
        doNothing().when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = (ReadinessStatus) ReflectionUtils.getField(f, probesService);
        assertNotNull(readinessStatus);
        assertFalse(readinessStatus.isReady());
    }

    @Test
    public void testTriggerReadinessChecksAccessServiceThrowsException() {
        when(artifactoryServers.getCurrentMember()).thenReturn(artifactoryServer);
        when(artifactoryServer.getServerState()).thenReturn(ArtifactoryServerState.RUNNING);
        when(artifactoryContext.isOffline()).thenReturn(false);
        when(artifactoryHome.getDataDir()).thenReturn(dataDir);
        when(dataDir.canWrite()).thenReturn(true);
        when(artifactoryHome.getLogDir()).thenReturn(logDir);
        when(logDir.canWrite()).thenReturn(true);

        doNothing().when(storageService).ping(true);
        doNothing().when(internalDbService).health();
        doThrow(new RuntimeException("What are you doing?")).when(accessService).ping();
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.triggerReadinessCheck();
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = (ReadinessStatus) ReflectionUtils.getField(f, probesService);
        assertNotNull(readinessStatus);
        assertFalse(readinessStatus.isReady());
    }

    @Test(dataProvider = "testCheckReadinessStatusNullDataProvider")
    public void testCheckReadinessStatusNull(boolean noCache) {
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        assertThrows(ServiceUnavailableException.class, () -> probesService.checkReadinessStatus(noCache));
    }

    @Test
    public void testCheckReadinessStatusReady() {
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = ReadinessStatus.builder()
                .running(true)
                .offline(false)
                .canWrite(true)
                .storageServiceReady(true)
                .internalDbServiceHealthy(true)
                .accessServiceReady(true)
                .build();
        ReflectionUtils.setField(f, probesService, readinessStatus);
        probesService.checkReadinessStatus(false);
        assertTrue(readinessStatus.isReady());
    }

    @Test(dataProvider = "testCheckReadinessStatusNotReadyDataProvider")
    public void testCheckReadinessStatusNotReady(boolean running, boolean offline, boolean canWrite,
            boolean storageServiceReady, boolean internalDbServiceHealthy, boolean accessServiceReady) {
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        ReadinessStatus readinessStatus = ReadinessStatus.builder()
                .running(running)
                .offline(offline)
                .canWrite(canWrite)
                .storageServiceReady(storageServiceReady)
                .internalDbServiceHealthy(internalDbServiceHealthy)
                .accessServiceReady(accessServiceReady)
                .build();
        ReflectionUtils.setField(f, probesService, readinessStatus);
        assertThrows(ServiceUnavailableException.class, () -> probesService.checkReadinessStatus(false));
        assertFalse(readinessStatus.isReady());
    }

    @Test
    public void testCheckLivenessStatusReadinessNull() {
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        String livenessFailReadinessEnabledKeyName = (String) ReflectionUtils.getField(f, probesService);
        when(artifactoryHome.getSysConfig()).thenReturn(sysConfig);
        when(sysConfig.getBoolean(livenessFailReadinessEnabledKeyName, true)).thenReturn(true);
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        assertThrows(ServiceUnavailableException.class, () -> probesService.checkLivenessStatus());
    }

    @Test
    public void testCheckLivenessStatusLivenessFailReadinessDisabled() {
        Field f = ReflectionUtils.findField(ProbesServiceImpl.class, "LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME");
        assertNotNull(f);
        ReflectionUtils.makeAccessible(f);
        String livenessFailReadinessEnabledKeyName = (String) ReflectionUtils.getField(f, probesService);
        when(artifactoryHome.getSysConfig()).thenReturn(sysConfig);
        when(sysConfig.getBoolean(livenessFailReadinessEnabledKeyName, true)).thenReturn(false);
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        probesService.checkLivenessStatus();
    }

    @Test
    public void testCheckLivenessStatusReadinessReady() {
        Field f1 = ReflectionUtils.findField(ProbesServiceImpl.class, "LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME");
        assertNotNull(f1);
        ReflectionUtils.makeAccessible(f1);
        String livenessFailReadinessEnabledKeyName = (String) ReflectionUtils.getField(f1, probesService);
        when(artifactoryHome.getSysConfig()).thenReturn(sysConfig);
        when(sysConfig.getBoolean(livenessFailReadinessEnabledKeyName, true)).thenReturn(true);
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        Field f2 = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f2);
        ReflectionUtils.makeAccessible(f2);
        ReadinessStatus readinessStatus = ReadinessStatus.builder()
                .running(true)
                .offline(false)
                .canWrite(true)
                .storageServiceReady(true)
                .internalDbServiceHealthy(true)
                .accessServiceReady(true)
                .build();
        ReflectionUtils.setField(f2, probesService, readinessStatus);
        probesService.checkLivenessStatus();
    }

    @Test
    public void testCheckLivenessStatusReadinessNotReady() {
        Field f1 = ReflectionUtils.findField(ProbesServiceImpl.class, "LIVENESS_FAIL_READINESS_ENABLED_KEY_NAME");
        assertNotNull(f1);
        ReflectionUtils.makeAccessible(f1);
        String livenessFailReadinessEnabledKeyName = (String) ReflectionUtils.getField(f1, probesService);
        when(artifactoryHome.getSysConfig()).thenReturn(sysConfig);
        when(sysConfig.getBoolean(livenessFailReadinessEnabledKeyName, true)).thenReturn(true);
        probesService = new ProbesServiceImpl(artifactoryServers, storageService, internalDbService, accessService,
                taskService);
        Field f2 = ReflectionUtils.findField(ProbesServiceImpl.class, "readinessStatus");
        assertNotNull(f2);
        ReflectionUtils.makeAccessible(f2);
        ReadinessStatus readinessStatus = ReadinessStatus.builder()
                .running(false)
                .offline(false)
                .canWrite(true)
                .storageServiceReady(true)
                .internalDbServiceHealthy(true)
                .accessServiceReady(true)
                .build();
        ReflectionUtils.setField(f2, probesService, readinessStatus);
        assertThrows(ServiceUnavailableException.class, () -> probesService.checkLivenessStatus());
    }

    @DataProvider
    public static Object[][] testTriggerReadinessChecksDataProvider() {
        return new Object[][] {
                // serverState, offline, canWriteDataDir, canWriteLogDir, expectedIsReady
                {ArtifactoryServerState.RUNNING, false, true, true, true},
                {ArtifactoryServerState.RUNNING, false, true, false, false},
                {ArtifactoryServerState.RUNNING, false, false, true, false},
                {ArtifactoryServerState.RUNNING, false, false, false, false},
                {ArtifactoryServerState.RUNNING, true, true, true, false},
                {ArtifactoryServerState.RUNNING, true, true, false, false},
                {ArtifactoryServerState.RUNNING, true, false, true, false},
                {ArtifactoryServerState.RUNNING, true, false, false, false},
                {ArtifactoryServerState.CONVERTING, false, true, true, false},
                {ArtifactoryServerState.CONVERTING, false, true, false, false},
                {ArtifactoryServerState.CONVERTING, false, false, true, false},
                {ArtifactoryServerState.CONVERTING, false, false, false, false},
                {ArtifactoryServerState.CONVERTING, true, true, true, false},
                {ArtifactoryServerState.CONVERTING, true, true, false, false},
                {ArtifactoryServerState.CONVERTING, true, false, true, false},
                {ArtifactoryServerState.CONVERTING, true, false, false, false}
        };
    }

    @DataProvider
    public static Object[][] testTriggerReadinessChecksLastSuccessfulReadinessZeroDataProvider() {
        return new Object[][] {
                // serverState, offline, canWriteDataDir, canWriteLogDir
                {ArtifactoryServerState.RUNNING, false, true, false},
                {ArtifactoryServerState.RUNNING, false, false, true},
                {ArtifactoryServerState.RUNNING, false, false, false},
                {ArtifactoryServerState.RUNNING, true, true, true},
                {ArtifactoryServerState.RUNNING, true, true, false},
                {ArtifactoryServerState.RUNNING, true, false, true},
                {ArtifactoryServerState.RUNNING, true, false, false},
                {ArtifactoryServerState.CONVERTING, false, true, true},
                {ArtifactoryServerState.CONVERTING, false, true, false},
                {ArtifactoryServerState.CONVERTING, false, false, true},
                {ArtifactoryServerState.CONVERTING, false, false, false},
                {ArtifactoryServerState.CONVERTING, true, true, true},
                {ArtifactoryServerState.CONVERTING, true, true, false},
                {ArtifactoryServerState.CONVERTING, true, false, true},
                {ArtifactoryServerState.CONVERTING, true, false, false}
        };
    }

    @DataProvider
    public static Object[][] testCheckReadinessStatusNullDataProvider() {
        return new Object[][] {{true}, {false}};
    }

    @DataProvider
    public static Object[][] testCheckReadinessStatusNotReadyDataProvider() {
        return new Object[][] {
                // running, offline, canWrite, storageServiceReady, internalDbServiceHealthy, accessServiceReady
                {true, true, true, true, true, true},
                {true, true, true, true, true, false},
                {true, true, true, true, false, true},
                {true, true, true, true, false, false},
                {true, true, true, false, true, true},
                {true, true, true, false, true, false},
                {true, true, true, false, false, true},
                {true, true, true, false, false, false},
                {true, true, false, true, true, true},
                {true, true, false, true, true, false},
                {true, true, false, true, false, true},
                {true, true, false, true, false, false},
                {true, true, false, false, true, true},
                {true, true, false, false, true, false},
                {true, true, false, false, false, true},
                {true, true, false, false, false, false},

                {true, false, true, true, true, false},
                {true, false, true, true, false, true},
                {true, false, true, true, false, false},
                {true, false, true, false, true, true},
                {true, false, true, false, true, false},
                {true, false, true, false, false, true},
                {true, false, true, false, false, false},
                {true, false, false, true, true, true},
                {true, false, false, true, true, false},
                {true, false, false, true, false, true},
                {true, false, false, true, false, false},
                {true, false, false, false, true, true},
                {true, false, false, false, true, false},
                {true, false, false, false, false, true},
                {true, false, false, false, false, false},
                {false, true, true, true, true, true},
                {false, true, true, true, true, false},
                {false, true, true, true, false, true},
                {false, true, true, true, false, false},
                {false, true, true, false, true, true},
                {false, true, true, false, true, false},
                {false, true, true, false, false, true},
                {false, true, true, false, false, false},
                {false, true, false, true, true, true},
                {false, true, false, true, true, false},
                {false, true, false, true, false, true},
                {false, true, false, true, false, false},
                {false, true, false, false, true, true},
                {false, true, false, false, true, false},
                {false, true, false, false, false, true},
                {false, true, false, false, false, false},
                {false, false, true, true, true, true},
                {false, false, true, true, true, false},
                {false, false, true, true, false, true},
                {false, false, true, true, false, false},
                {false, false, true, false, true, true},
                {false, false, true, false, true, false},
                {false, false, true, false, false, true},
                {false, false, true, false, false, false},
                {false, false, false, true, true, true},
                {false, false, false, true, true, false},
                {false, false, false, true, false, true},
                {false, false, false, true, false, false},
                {false, false, false, false, true, true},
                {false, false, false, false, true, false},
                {false, false, false, false, false, true},
                {false, false, false, false, false, false}
        };
    }
}
