package org.artifactory.core.api.ph.virtual;

import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageDownloadContext;
import org.artifactory.core.api.ph.resource.PackageUnfoundedResource;
import org.artifactory.core.api.ph.virtual.resolution.PackageRepo;
import org.artifactory.core.api.ph.virtual.resolution.PackageVirtualController;
import org.artifactory.core.api.ph.virtual.resolution.VirtualResolutionStrategyType;
import org.junit.jupiter.api.Assertions;
import org.testng.annotations.Test;

import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.when;

/**
 * Tests the Virtual repository visitor logic
 */
@Test
public class VirtualRepoTraversalResolverTest extends VirtualRepoTraversalBaseTest {

    /**
     * Test virtual repo resolution
     * 1. upload two artifacts into two different repositories that belongs to virtual repo
     * 2. try to download and expect to get first artifact
     */
    @Test
    public void testResolverExistTwoReturnFirst() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any());
        BaseResource result1 = resource(localRepoKey1, path, null);
        BaseResource result2 = resource(localRepoKey2, path, null);
        when(downloadService.getResource(localRepoKey1, path)).thenReturn(result1);
        when(downloadService.getResource(localRepoKey2, path)).thenReturn(result2);
        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource =
                virtualRepoTraversal.getInfo(new PackageDownloadContext(virtualRepoKey, path), controller);
        Assertions.assertEquals(localRepoKey1, resource.getRepoKey());
        Assertions.assertEquals(path, resource.getPath());
    }

    /**
     * Test virtual repo resolution
     * 1. upload two different artifacts into two different repositories that belongs to virtual repo
     * 2. try to download one of them and expect to find one artifact
     */
    @Test
    public void testExist1ReturnReturnFirst() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any());
        BaseResource result1 = resource(localRepoKey1, path1, null);
        BaseResource result2 = resource(localRepoKey2, path2, null);
        when(downloadService.getResource(localRepoKey1, path1)).thenReturn(result1);
        when(downloadService.getResource(localRepoKey2, path2)).thenReturn(result2);
        when(downloadService.getResource(localRepoKey1, path2))
                .thenReturn(new PackageUnfoundedResource(localRepoKey1, path2, 404, ""));
        when(downloadService.getResource(localRepoKey2, path1))
                .thenReturn(new PackageUnfoundedResource(localRepoKey2, path1, 404, ""));
        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource1 =
                virtualRepoTraversal.getInfo(new PackageDownloadContext(virtualRepoKey, path1), controller);
        Assertions.assertEquals(localRepoKey1, resource1.getRepoKey());
        Assertions.assertEquals(path1, resource1.getPath());
        BaseResource resource2 =
                virtualRepoTraversal.getInfo(new PackageDownloadContext(virtualRepoKey, path2), controller);
        Assertions.assertEquals(localRepoKey2, resource2.getRepoKey());
        Assertions.assertEquals(path2, resource2.getPath());
    }

    /**
     * Test virtual repo resolution
     * 1. upload two different artifacts into two different repositories that belongs to virtual repo
     * 2. try to download one of them and expect to find one artifact
     */
    @Test
    public void testNorFound() throws PackageException {
        List<? extends PackageRepo> subRepos = subRepositories(localRepoKey1, localRepoKey2);
        doAnswer(invocation -> subRepos).when(repoService).getSubRepositories(any());
        when(downloadService.getResource(localRepoKey1, path2))
                .thenReturn(new PackageUnfoundedResource(localRepoKey1, path2, 404, ""));
        when(downloadService.getResource(localRepoKey2, path1))
                .thenReturn(new PackageUnfoundedResource(localRepoKey2, path1, 404, ""));
        PackageVirtualController controller = controller(VirtualResolutionStrategyType.RESOLVER);
        BaseResource resource =
                virtualRepoTraversal.getInfo(new PackageDownloadContext(virtualRepoKey, path1), controller);
        Assertions.assertFalse(resource.isFound());
        Assertions.assertEquals(path1, resource.getPath());
    }
}
