/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.event.service;

import lombok.NonNull;
import org.artifactory.common.ConstantValues;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.db.event.dao.EventsDao;
import org.artifactory.storage.db.event.entity.EventRecord;
import org.artifactory.storage.db.event.entity.EventRecordHints;
import org.artifactory.storage.db.event.entity.EventsDaoParams;
import org.artifactory.storage.event.EventInfo;
import org.artifactory.storage.event.EventsService;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.ClockUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * A business service to interact with the events table.
 *
 * @author Yossi Shaul
 */
@Service
public class EventsServiceImpl implements EventsService, InternalEventsService {
    private static final Logger log = LoggerFactory.getLogger(EventsServiceImpl.class);
    /**
     * Optional file logger to print events to a file. Format of the file is timestamp|type|path
     */
    private static final Logger eventLog = LoggerFactory.getLogger("events");

    @Autowired
    private EventsDao eventsDao;

    @Autowired
    private EventsMetrics metrics;

    @Override
    public boolean isEnabled() {
        return ConstantValues.nodeEventsEnabled.getBoolean();
    }

    @Override
    public long getEventsCount() {
        try {
            return eventsDao.getEventsCount();
        } catch (SQLException e) {
            throw new StorageException("Failed to count events", e);
        }
    }

    @Override
    public void appendEvents(List<EventInfo> events) {
        List<EventRecord> records = new ArrayList<>();
        for (EventInfo event : events) {
            log.trace("Appending event {}", event);
            long timestamp = ClockUtils.epochMillis();
            records.add(toEventRecord(event, timestamp));
            metrics.event(event);
        }
        try {
            eventsDao.createBulk(records);
        } catch (SQLException e) {
            throw new StorageException("Failed to persist events", e);
        }

        logEvents(records);
    }

    @Override
    public List<EventInfo> getEventsAfter(long eventId, String repoKey, String fileSuffix, int limit) {
        try {
            List<EventRecord> records = eventsDao.loadEventsByParams(EventsDaoParams.builder()
                    .eventId(eventId)
                    .inclusive(false)
                    .limit(limit)
                    .repoKey(repoKey)
                    .suffix(fileSuffix)
                    .build());
            return records.stream().map(this::toEventInfo).collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve event", e);
        }
    }

    @Override
    public int countEventsAfter(long eventId, String repoKey, String fileSuffix) {
        try {
            return eventsDao.countEventsByParams(EventsDaoParams.builder()
                    .eventId(eventId)
                    .repoKey(repoKey)
                    .suffix(fileSuffix)
                    .build());
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve event", e);
        }
    }

    @Override
    public List<EventInfo> getEventsAfterInclusive(long timestamp, String repoKey) {
        try {
            List<EventRecord> records = eventsDao.loadNewerOrSameTimestamp(timestamp, repoKey);
            return records.stream().map(this::toEventInfo).collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve event", e);
        }
    }

    @Override
    public long getFirstEventTimestamp() {
        try {
            return eventsDao.getFirstEventTimestamp();
        } catch (SQLException e) {
            throw new StorageException("Failed to retrieve first event timestamp", e);
        }
    }

    @Override
    public List<EventInfo> getEventsByTimestamp(long timestamp) {
        try {
            List<EventRecord> records = eventsDao.loadEventsByTimestamp(timestamp);
            return records.stream().map(this::toEventInfo).collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Failed to get events by timestamp", e);
        }
    }

    @Override
    public Stream<EventInfo> getEventsSince(long timestamp, long maxEvents) {
        CompletableFuture<SQLException> onException = futureBounceSqlExceptionToStorageException();
        return eventsDao.loadNewerStream(timestamp, maxEvents, onException).map(this::toEventInfo);
    }

    @Override
    public List<EventInfo> getEventsInterval(long from, long to, int maxEvents, String repoKey) {
        CompletableFuture<SQLException> onException = futureBounceSqlExceptionToStorageException();
        try (Stream<EventRecord> eventRecordStream = eventsDao
                .loadIntervalStream(from, to, repoKey, maxEvents, onException)) {
            return eventRecordStream
                    .map(this::toEventInfo)
                    .collect(Collectors.toList());
        }
    }

    @Override
    public long getLastEventLogTimestamp() {
        try {
            return eventsDao.getLastEventTimestamp();
        } catch (SQLException e) {
            throw new StorageException("Failed to get last event log timestamp", e);
        }
    }

    @Override
    public long getFirstEventLogId(String repoKey) {
        try {
            return eventsDao.getFirstEventId(repoKey);
        } catch (SQLException e) {
            throw new StorageException("Failed to get first event id", e);
        }
    }

    @Override
    public long getLastEventLogId() {
        try {
            return eventsDao.getLastEventId();
        } catch (SQLException e) {
            throw new StorageException("Failed to get last event log id", e);
        }
    }

    private CompletableFuture<SQLException> futureBounceSqlExceptionToStorageException() {
        CompletableFuture<SQLException> onException = new CompletableFuture<>();
        onException.thenAccept(e -> {
            throw new StorageException("Failed to retrieve event", e);
        });
        return onException;
    }

    private EventInfo toEventInfo(EventRecord er) {
        return new EventInfo(er.getEventId(), er.getTimestamp(), er.getType(), er.getPath(),
                EventRecordHints.fromJson(er.getHints()));
    }

    private EventRecord toEventRecord(EventInfo e, long timestamp) {
        //passing the eventId here is not important, it is not used in Dao service
        return new EventRecord(-1, timestamp, e.getType(), e.getPath(),
                EventRecordHints.toJson(e.getHints()));
    }

    @Override
    public List<EventInfo> getAllEvents() {
        try {
            List<EventRecord> records = eventsDao.loadAll();
            return records.stream().map(this::toEventInfo).collect(Collectors.toList());
        } catch (SQLException e) {
            throw new StorageException("Failed to persist event", e);
        }
    }

    @Override
    public void deleteAll() {
        try {
            eventsDao.deleteAll();
        } catch (SQLException e) {
            throw new StorageException("Failed deleting all events", e);
        }
    }

    @Override
    public void deleteRange(long startInclusive, long endExclusive) {
        try {
            eventsDao.deleteRange(startInclusive, endExclusive);
        } catch (SQLException e) {
            throw new StorageException("Failed deleting all events older than " + startInclusive, e);
        }
    }

    /**
     * Shift tmp events from node_events_tmp to node_events
     * Delete node_events_tmp
     *
     * Note, this is a transactional method !
     *
     * @return repos key
     */
    @Override
    @Nonnull
    public Set<String> shiftEvents() {
        try {
            return eventsDao.shiftAndDeleteTmpEvents();
        } catch (Exception e) {
            throw new StorageException("Failed to shift events", e);
        }
    }

    @Override
    public int countEventsRangeByRepoKey(long startFrom, long end, @NotNull String repoKey) {
        try {
            return eventsDao.countRangeEventsByRepoPath(startFrom, end, repoKey);
        }catch (SQLException sqe){
            log.debug("Failed to get event range from {} to {} on repo {}. Reason: {}",
                    startFrom,
                    end,
                    repoKey,
                    sqe.getMessage());
            throw new StorageException("Failed to get event range on repo " + repoKey, sqe);
        }
    }

    private void logEvents(@NonNull List<EventRecord> events) {
        if (eventLog.isInfoEnabled()) {
            events.forEach(e -> eventLog.info("{}|{}|{}", e.getTimestamp(), e.getType(), e.getPath()));
        }
        if (log.isDebugEnabled()) {
            events.forEach(e -> log.debug("{}|{}|{}", e.getTimestamp(), e.getType(), e.getPath()));
        }
    }
}
